/*
 * @(#)SNMPport.java	1.00 4/23/01
 * Copyright (c) 2001 by Paul Regenhardt,  All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Paul Regenhardt ("Confidential Information"). You shall not 
 * disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Paul Regenhardt.
 */
import java.io.*;
import java.net.*;
import java.util.*;
import java.text.*;

/** This class is used to decode ASN messages.  Messages of this type are
 *   <ASN Object> = <type> <length> <ASN Object>
 * or
 *   <ASN Object> = <type> <length> <value>
 *
 * @author Paul Regenhardt
 * @version 1.0
 */

public class AsnObject extends Object 
{
  /** 
   * The type of object.
   */
  protected byte type;

  /** 
   * The starting position of the AsnObject in the SNMP byte sequence. 
   */
  protected int startPos = 0;

  /** 
   * The length of the header. That is the number of bytes of the 'type'
   * and 'length' fields. 
   */
  protected int headerLength = 0; 

  /** 
   * The length of the object.
   */
  protected int length = 0; 

    // classes
    public static final byte UNIVERSAL   =(byte)(0x00);
    public static final byte APPLICATION =(byte)(0x40);
    public static final byte CONTEXT     =(byte)(0x80);
    public static final byte PRIVATE     =(byte)(0xC0);

    // forms
    public static final byte PRIMITIVE   =(byte)(0x00);
    public static final byte CONSTRUCTOR =(byte)(0x20);

    // tags
    public static final byte BOOLEAN     =(byte)(UNIVERSAL | PRIMITIVE   | 0x01);
    public static final byte INTEGER     =(byte)(UNIVERSAL | PRIMITIVE   | 0x02);
    public static final byte BIT_STR     =(byte)(UNIVERSAL | PRIMITIVE   | 0x03);
    public static final byte OCTET_STR   =(byte)(UNIVERSAL | PRIMITIVE   | 0x04);
    public static final byte NULL        =(byte)(UNIVERSAL | PRIMITIVE   | 0x05);
    public static final byte OBJECT_ID   =(byte)(UNIVERSAL | PRIMITIVE   | 0x06);
    public static final byte SEQUENCE    =(byte)(UNIVERSAL | CONSTRUCTOR | 0x10);
    public static final byte SETOF       =(byte)(UNIVERSAL | CONSTRUCTOR | 0x11);

    public static final byte LONG_LEN    =(byte)(0x80);
    public static final byte EXTENSION_ID=(byte)(0x1F);
    public static final byte BIT8        =(byte)(0x80);

    public static final byte IPADDRESS    =(byte)(APPLICATION | PRIMITIVE  | 0);
    public static final byte COUNTER      =(byte)(APPLICATION | PRIMITIVE  | 1);
    public static final byte GAUGE        =(byte)(APPLICATION | PRIMITIVE  | 2);
    public static final byte TIMETICKS    =(byte)(APPLICATION | PRIMITIVE  | 3);
    public static final byte OPAQUE       =(byte)(APPLICATION | PRIMITIVE  | 4);
    public static final byte NSAP         =(byte)(APPLICATION | PRIMITIVE  | 5);
    public static final byte COUNTER64    =(byte)(APPLICATION | PRIMITIVE  | 6);
    public static final byte UINTEGER32   =(byte)(APPLICATION | PRIMITIVE  | 7);

    // Well known traps
    public static final byte COLDSTART         =(byte)(0x0);
    public static final byte WARMSTART         =(byte)(0x1);
    public static final byte LINKDOWN          =(byte)(0x2);
    public static final byte LINKUP            =(byte)(0x3);
    public static final byte AUTHFAIL          =(byte)(0x4);
    public static final byte EGPNEIGHBORLOSS   =(byte)(0x5);
    public static final byte ENTERPRISE        =(byte)(0x6);

    // Vars
    public static final byte NOSUCHOBJECT    =(byte)(CONTEXT | PRIMITIVE | 0x0);
    public static final byte NOSUCHINSTANCE  =(byte)(CONTEXT | PRIMITIVE | 0x1);
    public static final byte ENDOFMIBVIEW    =(byte)(CONTEXT | PRIMITIVE | 0x2);

    //errors
    public static final byte NOERROR            =(byte)(0x00);
    public static final byte TOOBIG             =(byte)(0x01);
    public static final byte NOSUCHNAME         =(byte)(0x02);
    public static final byte BADVALUE           =(byte)(0x03);
    public static final byte READONLY           =(byte)(0x04);
    public static final byte GENERR             =(byte)(0x05);
    public static final byte NOACCESS           =(byte)(0x06);
    public static final byte WRONGTYPE          =(byte)(0x07);
    public static final byte WRONGLENGTH        =(byte)(0x08);
    public static final byte WRONGENCODING      =(byte)(0x09);
    public static final byte WRONGVALUE         =(byte)(0x0A);
    public static final byte NOCREATION         =(byte)(0x0B);
    public static final byte INCONSISTENTVALUE  =(byte)(0x0C);
    public static final byte RESOURCEUNAVAILABLE=(byte)(0x0D);
    public static final byte COMMITFAILED       =(byte)(0x0E);
    public static final byte UNDOFAILED         =(byte)(0x0F);
    public static final byte AUTHORIZATIONERR   =(byte)(0x10);
    public static final byte NOTWRITABLE        =(byte)(0x11);
    public static final byte INCONSISTENTNAME   =(byte)(0x12);
    public static final byte DECODING_EXC       =(byte)(0x13);

    public static final byte VERSION_1              =(byte)(0x0);
    public static final byte VERSION_2c             =(byte)(0x1);
    public static final byte VERSION_3              =(byte)(0x3);

    public static final byte GET_REQ_MSG     =(byte)(CONTEXT | CONSTRUCTOR | 0x0);
    public static final byte GETNEXT_REQ_MSG =(byte)(CONTEXT | CONSTRUCTOR | 0x1);
    public static final byte GET_RSP_MSG     =(byte)(CONTEXT | CONSTRUCTOR | 0x2);
    public static final byte SET_REQ_MSG     =(byte)(CONTEXT | CONSTRUCTOR | 0x3);
    public static final byte TRP_REQ_MSG     =(byte)(CONTEXT | CONSTRUCTOR | 0x4);

    public static final byte GETBULK_REQ_MSG =(byte)(CONTEXT  | CONSTRUCTOR | 0x5);
    public static final byte INFORM_REQ_MSG  =(byte)(CONTEXT  | CONSTRUCTOR | 0x6);
    public static final byte TRPV2_REQ_MSG   =(byte)(CONTEXT  | CONSTRUCTOR | 0x7);
    public static final byte GET_RPRT_MSG    =(byte)(CONTEXT  | CONSTRUCTOR | 0x8);

    protected Vector contents = null;
    protected String sContents = null;
    protected Integer iValue = null;
    protected Long  lValue = null;
    protected byte[] data = null;
    
    private int pos = 0;
    
    private byte getType() {return type;}
    private void skip(int i) {pos += i;}
    private void reset()     {pos = startPos;}
    protected int totalLength()   {return length+headerLength;}
    private byte read() throws Exception
    {
        if (pos >= data.length) 
            throw new Exception("Attempting to read beyond buffer");
        return data[startPos + pos++];
    }
    
    // default
    AsnObject()
    {
        type = 0;
        pos = 0;
        startPos = 0;
        length = -1;
        headerLength = -1;
    }
    
    // This constructor starts the chain of events, at this point the type
    // isn't known
    AsnObject(byte[] ibuf) throws Exception
    {
        super();
        data =  ibuf;
        marshallObject();
    }
    
    // Create a new AsnObject of type t, this is the recursive constructor
    AsnObject(int p, byte[] ibuf) throws Exception
    {
        super();
        startPos = p;
        data = ibuf;
        marshallObject();
    }
    
    // load the current object with objects
    public AsnObject marshallObject() throws Exception
    {
        int cp = pos;
        int i;
        int val; 
        
        type = read();
        length  = getLength();
        headerLength = pos - cp; // headerlength in bytes

        // if it's an object we recursively build it
        switch (type)
        {
          case SEQUENCE:
          case GET_RSP_MSG :
          case GET_RPRT_MSG :
          case TRPV2_REQ_MSG :
          case TRP_REQ_MSG :
            contents = new Vector();
            while (pos < length) {
                AsnObject ao = new AsnObject(pos+startPos, data);
                contents.addElement(ao);
                pos += ao.totalLength();
            }
            break;
          case UINTEGER32 :
          case TIMETICKS :
          case COUNTER :
          case GAUGE :
            if ( length > 4 ) throw new Exception ("Exception: Bad Length of 4 byte counter, to many bytes");
            val = 0;
            for (i=0; i < length; i++){
                byte n = read();
                int v = (int) (0x7f & n);
                if (n < 0) v+=128;
                val = (val*256) + v;
            }
            iValue = new Integer(val);
            break;
          case COUNTER64 :
            long lv = 0;
            if ( length > 8 ) throw new Exception ("Exception: Bad Length of Counter64");
            for (i=0; i < length; i++){
                byte n = read();
                int v = (int) (0x7f & n);
                if (n < 0) v+=128;
                lv = (lv*256) + v;
            }
            lValue = new Long(lv);
            break;
          case INTEGER :
            val = 0;
            if ( length > 4 ) throw new Exception ("Exception: Bad Length of 4 byte counter, to many bytes");
            for (i=0; i < length; i++){
                byte n = read();
                int v = (int) (0x7f & n);
                if (n < 0) v+=128;
                val = (val*256) + v;
            }
            iValue = new Integer(val);
            break;
          case OPAQUE :
          case NSAP :
          case IPADDRESS :
            contents = new Vector();
            for (i=0; i < length; i++){
                byte n = read();
                int v = (int) (0x7f & n);
                if (n < 0) v+=128;
                contents.addElement(new Integer(v));
            }
            break;
          
          case OBJECT_ID :
            contents = new Vector();
            byte n = read();
            int v = n / 40; 
            contents.addElement(new Integer(v));
            v = n % 40; 
            contents.addElement(new Integer(v));
            
            val = 0;
            for (i=1; i < length; i++){
                n = read();
                v = (int) (0x7f & n);
                val = val << 7;
                val |= v;
                if ( n >= 0) {
                    contents.addElement(new Integer(val));
                    val = 0;
                }
            }
            break;
          case OCTET_STR :
            byte[] r = new byte[length];
            for (i = 0; i<length; i++) r[i] = read();
            sContents = new String(r);
            break;
          case NULL :
            return null;
          case NOSUCHOBJECT:
          case NOSUCHINSTANCE:
          case ENDOFMIBVIEW:
            pos += length;
            break;
          default :
            pos += length;
            System.out.print("\nunknown type: ");
            System.out.print(type);
            break;
        }
        return this;
    }
    // handle simple and complex lengths
    int getLength() throws Exception
    {
        int length = 0;
        byte len = read();

        if ((0x80 & len) != 0 ) // complex
        {
            int count = (0x7f & len);
            if ( count > 4 ) throw new Exception("Length of object great than 4 bytes");
            for (int i=0; i < count; i++){
                byte b = read();
                int n = 0x7f & b;
                if ((0x80 & b) != 0) n += 128;
                length = (length << 8) + n;
            }
        } 
        // simple
        else length = (int) (0x7f & len);
        
        return (length);
    }
    
    public AsnObject getObject(int position) throws Exception
    {
        if (contents == null) throw new Exception("Not a Container Object, can not be indexed");
        if (position > contents.size()) throw new Exception("Index larger than container");
        AsnObject ao = (AsnObject) contents.elementAt(position);
        return ao;
    }
    
    public boolean isInt()
    {
        return (type == INTEGER     || 
                type == UINTEGER32  || 
                type == TIMETICKS   || 
                type == COUNTER     || 
                type == GAUGE);
    }

    // get an integer object from the current object
    public int getInt() throws Exception
    {
        if (!isInt()) throw new Exception("Object is not an integer");
        return iValue.intValue();
    }
    public int getInt(int position) throws Exception
    {
        if (contents == null) throw new Exception("Not a Container Object, can not be indexed");
        if (position > contents.size()) throw new Exception("Index larger than container");
        AsnObject ao = (AsnObject) contents.elementAt(position);
        if (ao == null) throw new Exception("Object was null");
        return ao.getInt();
    }
        public long getLong() throws Exception
    {
        if (type != COUNTER64) throw new Exception("Object is not an integer");
        return lValue.longValue();
    }

    public long getLong(int position) throws Exception
    {
        if (contents == null) throw new Exception("Not a Container Object, can not be indexed");
        if (position > contents.size()) throw new Exception("Index larger than container");
        AsnObject ao = (AsnObject) contents.elementAt(position);
        if (ao == null) throw new Exception("Object was null");
        return ao.getLong();
    }

    // get a String object from the current object
    public String getString() throws Exception
    {
        if (type != OCTET_STR) throw new Exception("Object is not an integer");
        return sContents;
    }
    public String getString(int position) throws Exception
    {
        if (contents == null) throw new Exception("Not a Container Object, can not be indexed");
        if (position > contents.size()) throw new Exception("Index larger than container");
        AsnObject ao = (AsnObject) contents.elementAt(position);
        if (ao == null) throw new Exception("Object was null");
        return ao.getString();
    }
    public void print(PrintStream p, int offset)
    {
        int i=0;
        String leading = new String();
        for (i=0; i < offset; i++) leading = leading + " ";
        
        switch (type)
        {
          case SEQUENCE:
            p.println(leading + "Sequence");
            for (i=0; i < contents.size(); i++) {
                AsnObject ao = (AsnObject) contents.elementAt(i);
                ao.print(p, offset+3);
            }
            break;
          case GET_RSP_MSG :
            p.println(leading + "GET_RSP_MSG");
            for (i=0; i < contents.size(); i++) {
                AsnObject ao = (AsnObject) contents.elementAt(i);
                ao.print(p, offset+3);
            }
            break;
          case GET_RPRT_MSG :
            p.println(leading + "GET_RPRT_MSG");
            for (i=0; i < contents.size(); i++) {
                AsnObject ao = (AsnObject) contents.elementAt(i);
                ao.print(p, offset+3);
            }
            break;
          case TRPV2_REQ_MSG :
            p.println(leading + "TRPV2_REQ_MSG");
            for (i=0; i < contents.size(); i++) {
                AsnObject ao = (AsnObject) contents.elementAt(i);
                ao.print(p, offset+3);
            }
            break;
          case TRP_REQ_MSG :
            p.println(leading + "TRP_REQ_MSG");
            for (i=0; i < contents.size(); i++) {
                AsnObject ao = (AsnObject) contents.elementAt(i);
                ao.print(p, offset+3);
            }
            break;
          case UINTEGER32 :
            p.println(leading + "UINTEGER32 "+ iValue);
            break;
          case TIMETICKS :
	        SimpleDateFormat dFormat = new SimpleDateFormat ("MM/dd/yy  HH:mm:ss");
	        long ld = iValue.longValue();
            String d = dFormat.format(new Date(ld*1000)); 
            p.println(leading + "TIMETICKS "+ d);
            break;
          case COUNTER :
            p.println(leading + "COUNTER "+ iValue);
            break;
          case GAUGE :
            p.println(leading + "GAUGE "+ iValue);
            break;
          case COUNTER64 :
            p.println(leading + "COUNTER64 "+ lValue);
            break;
          case INTEGER :
            p.println(leading + "INTEGER "+ iValue);
            break;
          case OPAQUE :
            p.print(leading + "OPAQUE ");
            for (i=0; i < length; i++){
                Integer iv = (Integer) contents.elementAt(i);
                p.print(" "+iv);
            }
            p.println("");
            break;
          case NSAP :
            p.print(leading + "NSAP ");
            for (i=0; i < length; i++){
                Integer iv = (Integer) contents.elementAt(i);
                p.print(" "+iv);
            }
            p.println("");
            break;
          case IPADDRESS :
            p.print(leading + "IPADDRESS ");
            for (i=0; i < length; i++){
                Integer iv = (Integer) contents.elementAt(i);
                p.print(" "+iv);
            }
            p.println("");
            break;
          case OBJECT_ID :
            p.print(leading + "OBJECT_ID ");
            for (i=0; i < contents.size(); i++){
                Integer iv = (Integer) contents.elementAt(i);
                p.print(" "+iv);
            }
            p.println("");
            break;
          case OCTET_STR :
            p.println(leading + "String: "+sContents);
            break;
          case NULL :
            p.println(leading + "NULL ");
            break;
          default :
            p.print("unknown Object, type: ");
            p.println(type);
            break;
        }
    }
}


